import subprocess
from autogradercpp.file_handler import search_cpp_files, read_test_cases
from autogradercpp.diff import find_and_print_differences
from autogradercpp import ansi_colors
from autogradercpp.exceptions import GccCompilerException

class AutograderCpp:

    """
    List of supported run_modes (case sensitive):
    - verbose
    - standard
    - concise
    - summary

    """

    def __init__(self, run_mode:str="standard", timeout_sec:float=5, base_path:str="") -> None:
        self.configure(run_mode=run_mode, timeout_sec=timeout_sec, base_path=base_path)
        if not self.check_gcc_installed():
            raise GccCompilerException("Could not detect an installed instance of gcc/g++. Please check the compiler and try again.")

    def configure(self, run_mode:str="standard", timeout_sec:float=5, base_path:str="") -> None:
        self.run_mode = run_mode
        self.timeout_sec = timeout_sec
        self.base_path=base_path

    def check_gcc_installed(self):
        try:
            subprocess.run(
                ['g++', '--version'], 
                stdout=subprocess.PIPE, 
                stderr=subprocess.PIPE, 
                check=True
                )
            return True
        except subprocess.CalledProcessError:
            return False

    def run_test(self,input_str:str, expected_output:str, question_path:str) -> bool: 
        # If no output found, wait for timeout_sec (default 5) seconds before throwing a timeout and moving on. This ensures program moves on even if DUT is stuck in infinite loop

        # Compile the C++ code
        compile_command = "g++ " + str(question_path)
        # print(compile_command)
        subprocess.run(compile_command, shell=True, stdout=subprocess.PIPE, stderr=subprocess.PIPE)
        # Run the compiled program with input
        run_command = f"./a.out"

        try:
            process = subprocess.run(run_command, input=input_str, shell=True, stdout=subprocess.PIPE, stderr=subprocess.PIPE, text=True, timeout = self.timeout_sec)
            # Check if the output matches the expected output
            actual_output = process.stdout.strip()
            passFlag =  actual_output == expected_output.strip()
            if (self.run_mode=="standard" and not passFlag) or self.run_mode=="verbose":
                find_and_print_differences(actual_output,expected_output)
            return passFlag
        except subprocess.TimeoutExpired:
            if self.run_mode!= "summary":
                print(ansi_colors.colorize_line(f"Test case timed out after {self.timeout_sec} seconds.",ansi_colors.ERROR_COLOR))
            return False

    def autograde(self, path:str, input_test_cases:str, output_test_cases:str) -> None:
        enumerated_test_cases = enumerate(zip(input_test_cases, output_test_cases), 1)
        total_cases = len(input_test_cases)
        num_test_passed = 0
        for i, (input_str, expected_output) in enumerated_test_cases:
            try:
                if self.run_mode != "summary":
                    print(ansi_colors.colorize_line(f"\nEvaluating test case {i}\n",ansi_colors.FOREGROUND_COLORS['underline']), end='' if self.run_mode == "concise" else '\n')
                testSuccess = self.run_test(input_str=input_str, expected_output=expected_output, question_path=path) 
                num_test_passed+=testSuccess
            except Exception as e:
                testSuccess = False
                if self.run_mode != "summary":
                    print(ansi_colors.colorize_line("An error occured while evaluting the test case", ansi_colors.ERROR_COLOR))
                    print(e)
            if self.run_mode != "summary":
                if testSuccess:
                    print(ansi_colors.colorize_line(f"Test case {i}: Passed", ansi_colors.TEST_PASSED_COLOR))
                else:
                    print(f"Test case {i}: Failed")
                if self.run_mode != "concise":
                    print()
        print(ansi_colors.colorize_line(f"Passed {num_test_passed}/{total_cases}",ansi_colors.BACKGROUND_COLORS['blue']))
            
    def grade_root_dir(self, ROOT_FOLDER_PATH:str, FILENAME_TO_MATCH:str, TEST_INPUT_FILE_PATH:str, TEST_OUTPUT_FILE_PATH:str) -> None:
        file_paths = search_cpp_files(self.base_path + ROOT_FOLDER_PATH, FILENAME_TO_MATCH)
        input_test_cases = read_test_cases(self.base_path + TEST_INPUT_FILE_PATH)
        output_test_cases = read_test_cases(self.base_path + TEST_OUTPUT_FILE_PATH)
        for roll in file_paths:
            print('\n'+"-"*100)
            path = file_paths[roll]
            message =ansi_colors.colorize_line( ansi_colors.colorize_line(f"{roll}: {path if path == 'absent' else 'present'}", ansi_colors.FOREGROUND_COLORS['yellow']), ansi_colors.FOREGROUND_COLORS['bold'])
            print(message)
            if path !="absent":
                self.autograde(path, input_test_cases, output_test_cases)


from pathlib import Path
from autogradercpp.exceptions import TestBenchValidationException

def search_cpp_files(root_folder:str, filename_to_match:str) -> dict:
    """The given function expects that there is a directory of following order -

        root_folder/\n
            |-- roll1/\n
            | |---- file1.cpp\n
            | |---- file2.cpp\n
            | |---- some_filename_to_match.cpp\n
            |--roll2/\n
            | |---- file1.cpp\n
            ...
    
    roll is the unique identification string of the submitter 
    
    The function looks for every folder/roll in the root folder and checks for the first file with filename containing the keyword filename_to_match. If found, the path to it is returned along with the roll of the person. If for a given roll no such file found, the function returns absent corresponding to the path for the roll.
    Args:
        root_folder (str): path to the directory containing folders and files under test
        filename_to_match (str): the unique keyword/pattern in file names

    Returns:
        dict: key: name of folder/roll: path to file
    """
    root = Path(root_folder)
    cpp_file_path_dict={}
    for roll_number_path in root.iterdir():
        if roll_number_path.is_dir():
            found = False
            for file_path in roll_number_path.rglob('*.cpp'):
                if filename_to_match in file_path.name:
                    cpp_file_path_dict[roll_number_path.name]=file_path
                    found = True
                    break
            if not found:
                cpp_file_path_dict[roll_number_path.name]="absent"
    return cpp_file_path_dict

def read_test_cases(file_path:str) -> list[str]:
    file = open(file_path, "r") 
    lines = file.readlines()
    test_cases = []
    file_category = None
    for line in lines:
        if line.startswith("#testcase_input"):
            if file_category == "output":
                raise TestBenchValidationException("Input values found in output test case file, please check the test files")
            test_cases.append("")
            if file_category == None:
                file_category = "input"
            
        elif line.startswith("#testcase_output"):
            if file_category == "input":
                raise TestBenchValidationException("Output values found in input test case file, please check the test files")
            test_cases.append("")
            if file_category == None:
                file_category = "output"
        else:
            test_cases[-1] += line
    file.close()
    return test_cases



FOREGROUND_COLORS={
    "black":30,
    "red":31,
    "green":32,
    "yellow":33,
    "blue":34,
    "magenta":35,
    "cyan":36,
    "white":37,
    "default":39,
    "reset":0,
    "bright_black":90,
    "bright_red":91,
    "bright_green":92,
    "bright_yellow":93,
    "bright_blue":94,
    "bright_magenta":95,
    "bright_cyan":96,
    "bright_white":97,
    "bold":1,
    "italic":3,
    "underline":4,
}
BACKGROUND_COLORS={
    "black":40,
    "red":41,
    "green":42,
    "yellow":43,
    "blue":44,
    "magenta":45,
    "cyan":46,
    "white":47,
    "default":49,
    "reset":0,
    "bright_black":100,
    "bright_red":101,
    "bright_green":102,
    "bright_yellow":103,
    "bright_blue":104,
    "bright_magenta":105,
    "bright_cyan":106,
    "bright_white":107,
}
ERROR_COLOR = BACKGROUND_COLORS['red']
TEST_EVAL_COLOR = FOREGROUND_COLORS['yellow']
TEXT_EXPECTED_COLOR = FOREGROUND_COLORS['bright_green']
TEXT_FOUND_COLOR = FOREGROUND_COLORS['bright_red']
TEST_PASSED_COLOR=BACKGROUND_COLORS['green']
TEST_FAILED_COLOR= BACKGROUND_COLORS['magenta']

def colorize_line(line, color_code):
    return f"\033[{color_code}m{line}\033[0m"



from autogradercpp.ansi_colors import TEXT_FOUND_COLOR, TEXT_EXPECTED_COLOR, colorize_line

def find_and_print_differences(actual_text:str, expected_text:str) -> None:
    lines_actual = actual_text.splitlines()
    lines_expected = expected_text.splitlines()

    max_len = max(len(lines_actual), len(lines_expected))
    
    print("Line   | Actual Output        | Expected Output")
    print("-------|----------------------|---------------------")

    for i, (line_actual, line_expected) in enumerate(zip(lines_actual + [''] * (max_len - len(lines_actual)), lines_expected + [''] * (max_len - len(lines_expected))), 1):
        if line_actual != line_expected:
            line_actual_colored = colorize_line(line_actual, TEXT_FOUND_COLOR) 
            line_expected_colored = colorize_line(line_expected, TEXT_EXPECTED_COLOR) 
            print(f"{i:<7}| {line_actual_colored.ljust(29)} | {line_expected_colored}")
        else:
            print(f"{i:<7}| {line_actual.ljust(20)} | {line_expected}")


from autogradercpp import ansi_colors
class TestBenchValidationException(Exception):
    def __init__(self, message:str) -> None:
        self.message = message
    def __str__(self) -> str:
        return ansi_colors.colorize_line(self.message, ansi_colors.ERROR_COLOR)

class GccCompilerException(Exception):
    def __init__(self, message:str) -> None:
        self.message = message
    def __str__(self) -> str:
        return ansi_colors.colorize_line(self.message, ansi_colors.ERROR_COLOR)



from setuptools import find_packages, setup

# read the contents of your README file
from pathlib import Path
this_directory = Path(__file__).parent
long_description = (this_directory / "README.md").read_text()

setup(
    name='autogradercpp',
    packages=find_packages(
        include=['autogradercpp']
    ),
    version='0.1.3',
    description='An autograder library for evaluating C++ programs.',
    author='Abhijat Bharadwaj (Keymii)',
    author_email='bharadwaj.abhijat@yahoo.com',
    classifiers=[
        'Programming Language :: Python :: 3',
        'License :: OSI Approved :: MIT License',
        'Operating System :: OS Independent',
    ],
    install_requires=[],
    long_description=long_description,
    long_description_content_type='text/markdown',
    project_urls = {
        'GitHub':'https://github.com/Keymii/autograder-for-cpp'
    }
)
